(* signature for device drivers and interrupt handlers *)

signature DEV_INT = sig
    exception INVALID_INTERRUPT_ID
    exception NO_ASSOCIATED_INTERRUPT_HANDLER
    type handler
    datatype dev_status = UNINSTALLED | INSTALLED of handler 
    type dev_info
    type IRQ = int
    type priority

    val KEYBOARD_PRIORITY: priority
    val TIMER_PRIORITY: priority
    val NET_PRIORITY: priority
    
    val listHandlers: unit -> unit
    val setHandler: (IRQ * dev_status * priority * string) -> dev_status
    val clearHandler: IRQ  -> unit
    val suspend_interrupts: unit -> unit
    val resume_interrupts: unit -> unit
    val mask: IRQ -> unit	(* interrupt handler should not call this *)
    val unmask: IRQ -> unit	(* interrupt handler should not call this *)
    val dispatch: IRQ -> unit
    val init: unit -> unit 
end  
