(* device driver for Ethernet Adaptor *)
(* translated to SML and adapted from Linux Debian version 2.0.33
   file: linux/if_ether.h
   GF 2/1999
 *)

structure EthConstant : ETHCONSTANT = struct

(* following constant value came from linux/if_ether.h *)

(*
 *      IEEE 802.3 Ethernet magic constants.  The frame sizes omit the preamble
 *      and FCS/CRC (frame check sequence).
 *)
val ETH_ALEN =        6		(* Octets in one ethernet addr   *)
val ETH_HLEN =       14        	(* Total octets in header.       *)
val ETH_ZLEN =       60         (* Min. octets in frame sans FCS *)
val ETH_DATA_LEN =   1500       (* Max. octets in payload        *)  
val ETH_FRAME_LEN =  1514       (* Max. octets in frame sans FCS *)

(*
 *	These are the defined Ethernet Protocol ID's.
 *)

val ETH_P_LOOP	= 0x0060		(* Ethernet Loopback packet	*)
val ETH_P_ECHO	= 0x0200		(* Ethernet Echo packet		*)
val ETH_P_PUP	= 0x0400		(* Xerox PUP packet		*)
val ETH_P_IP	= 0x0800		(* Internet Protocol packet	*)
val ETH_P_X25	= 0x0805		(* CCITT X.25			*)
val ETH_P_ARP	= 0x0806		(* Address Resolution packet	*)
val ETH_P_BPQ	= 0x08FF		(* G8BPQ AX.25 Ethernet Packet	
				   	   [ NOT AN OFFICIALLY REGISTERED ID ] *)
val ETH_P_DEC      = 0x6000          	(* DEC Assigned proto           *)
val ETH_P_DNA_DL   = 0x6001          	(* DEC DNA Dump/Load            *)
val ETH_P_DNA_RC   = 0x6002          	(* DEC DNA Remote Console       *)
val ETH_P_DNA_RT   = 0x6003             (* DEC DNA Routing              *)
val ETH_P_LAT      = 0x6004          	(* DEC LAT                      *)
val ETH_P_DIAG     = 0x6005          	(* DEC Diagnostics              *)
val ETH_P_CUST     = 0x6006          	(* DEC Customer use             *)
val ETH_P_SCA      = 0x6007          	(* DEC Systems Comms Arch       *)
val ETH_P_RARP     = 0x8035	     	(* Reverse Addr Res packet	*)
val ETH_P_ATALK	   = 0x809B	     	(* Appletalk DDP		*)
val ETH_P_AARP	   = 0x80F3	     	(* Appletalk AARP		*)
val ETH_P_IPX	   = 0x8137	     	(* IPX over DIX			*)
val ETH_P_IPV6	   = 0x86DD	     	(* IPv6 over bluebook		*)

(*
 *	Non DIX types. Won't clash for 1500 types.
 *)
 
val ETH_P_802_3	= 0x0001		(* Dummy type for 802.3 frames  *)
val ETH_P_AX25	= 0x0002		(* Dummy protocol id for AX.25  *)
val ETH_P_ALL	= 0x0003		(* Every packet (be careful!!!) *)
val ETH_P_802_2	= 0x0004		(* 802.2 frames 		*)
val ETH_P_SNAP	= 0x0005		(* Internal only		*)
val ETH_P_DDCMP    = 0x0006          	(* DEC DDCMP: Internal only     *)
val ETH_P_WAN_PPP  = 0x0007          	(* Dummy type for WAN PPP frames*)
val ETH_P_PPP_MP   = 0x0008          	(* Dummy type for PPP MP frames *)
val ETH_P_LOCALTALK= 0x0009		(* Localtalk pseudo type 	*)
val ETH_P_PPPTALK  = 0x0010		(* Dummy type for Atalk over PPP*)
val ETH_P_TR_802_2 = 0x0011		(* 802.2 frames 		*)

(*
 *	This is an Ethernet frame header.
 *)

end

