(* device driver for the network adaptor *)

(* translated to SML and adapted from Linux Debian version 2.0.33
   files: 
          linux/if.h

   GF 2/1999
 *)

structure NetDevConstant : NETDEVCONSTANT = struct

(* following constant value came from linux/if.h *)
(* Standard interface flags. *)
val	IFF_UP		= 0wx1		(* interface is up		*)
val	IFF_BROADCAST	= 0wx2		(* broadcast address valid	*)
val	IFF_DEBUG	= 0wx4		(* turn on debugging		*)
val	IFF_LOOPBACK	= 0wx8		(* is a loopback net		*)
val	IFF_POINTOPOINT	= 0wx10		(* interface is has p-p link	*)
val	IFF_NOTRAILERS	= 0wx20		(* avoid use of trailers	*)
val	IFF_RUNNING	= 0wx40		(* resources allocated		*)
val	IFF_NOARP	= 0wx80		(* no ARP protocol		*)
val	IFF_PROMISC	= 0wx100	(* receive all packets		*)
(* Not supported *)
val	IFF_ALLMULTI	= 0wx200	(* receive all multicast packets*)

val IFF_MASTER		= 0wx400	(* master of a load balancer 	*)
val IFF_SLAVE		= 0wx800	(* slave of a load balancer	*)

val IFF_MULTICAST	= 0wx1000	(* Supports multicast		*)
val IFF_SOFTHEADERS 	= 0wx2000	(* Device cannot construct headers
					 * until broadcast time. Therefore
					 * SOCK_PACKET must call header
					 * construction. Private flag.
					 * Never visible outside of kernel.
					 *)

(* following constant came from linux/if_arp.h *)
(* ARP protocol HARDWARE identifiers. *)
val ARPHRD_NETROM =	0		(* from KA9Q: NET/ROM pseudo	*)
val ARPHRD_ETHER  =	1		(* Ethernet 10Mbps		*)
val ARPHRD_EETHER =	2		(* Experimental Ethernet	*)
val ARPHRD_AX25   =	3		(* AX.25 Level 2		*)
val ARPHRD_PRONET =	4		(* PROnet token ring		*)
val ARPHRD_CHAOS  =	5		(* Chaosnet			*)
val ARPHRD_IEEE802=	6		(* IEEE 802.2 Ethernet/TR/TB	*)
val ARPHRD_ARCNET =	7		(* ARCnet			*)
val ARPHRD_APPLETLK=	8		(* APPLEtalk			*)
val ARPHRD_DLCI	  = 	15		(* Frame Relay DLCI		*)
val ARPHRD_METRICOM=	23		(* Metricom STRIP (new IANA id)	*)

(* Dummy types for non ARP hardware *)
val ARPHRD_SLIP	   =	256
val ARPHRD_CSLIP   =	257
val ARPHRD_SLIP6   =	258
val ARPHRD_CSLIP6  =	259
val ARPHRD_RSRVD   =	260		(* Notional KISS type 		*)
val ARPHRD_ADAPT   =	264
val ARPHRD_PPP	   =	512

val ARPHRD_TUNNEL  =	768		(* IPIP tunnel			*)
val ARPHRD_TUNNEL6 =	769		(* IPIP6 tunnel			*)
val ARPHRD_FRAD	   =	770		(* Frame Relay Access Device	*)
val ARPHRD_SKIP	   =	771		(* SKIP vif			*)
val ARPHRD_LOOPBACK=	772		(* Loopback device		*)
val ARPHRD_LOCALTLK=    773		(* Localtalk device		*)
val ARPHRD_FDDI	   =	774		(* Fiber Distributed Data Interface *)

(* ARP protocol opcodes. *)
val ARPOP_REQUEST  =	1		(* ARP request			*)
val ARPOP_REPLY    =	2		(* ARP reply			*)
val ARPOP_RREQUEST =	3		(* RARP request			*)
val ARPOP_RREPLY   =	4		(* RARP reply			*)

(* following constant came from linux/socket.h *)
(* Supported address families. *)
val AF_UNSPEC	=	0
val AF_UNIX	=	1	(* Unix domain sockets 		*)
val AF_INET	=	2	(* Internet IP Protocol 	*)
val AF_AX25	=	3	(* Amateur Radio AX.25 		*)
val AF_IPX	=	4	(* Novell IPX 			*)
val AF_APPLETALK=	5	(* Appletalk DDP 		*)
val AF_NETROM	=	6	(* Amateur radio NetROM 	*)
val AF_BRIDGE	=	7	(* Multiprotocol bridge 	*)
val AF_AAL5	=	8	(* Reserved for Werner's ATM 	*)
val AF_X25	=	9	(* Reserved for X.25 project 	*)

(* Protocol families, same as address families. *)
val PF_UNSPEC	=	AF_UNSPEC
val PF_UNIX	=	AF_UNIX
val PF_INET	=	AF_INET
val PF_AX25	=	AF_AX25
val PF_IPX	=	AF_IPX
val PF_APPLETALK=	AF_APPLETALK
val PF_NETROM	= 	AF_NETROM
val PF_BRIDGE	=	AF_BRIDGE
val PF_AAL5	=	AF_AAL5
val PF_X25	=	AF_X25


end










