(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Chris Stone (stone+@cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15213-3891

		i.	Abstract

        Features new in the 2.0 release:
          --- correctly serves compressed files
          --- can redirect URLs starting with ~ to the "people" subdirectory
              (that is, ...cs.cmu.edu/~cstone/foo is equivalent to the URL
                        ...cs.cmu.edu/people/cstone/foo)
          --- Server actions are more easily reconfigurable
              (it's very easy to define aliases such as ~ above, or
               to allow documents to be served from a directory tree
               rooted somewhere else than /afs/cs/project/fox/mosaic)          



		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Httpd
	2.	translation routines
	3.	Act upon requests
	4.	"server-side includes" (sort of)
	5.	parsing functions
	6.	packet-parsing
	7.	HTTP/HTML output
	8.	request logging
	9.	internal function process_request
	10.	function install
	11.	functions for reporting statistics
	12.	inetd_service definition

		iii.	RCS Log
	
$Log:	httpd.fun,v $
Revision 1.26  97/12/11  20:03:50  esb
brought up to speed with the new posix interface.

Revision 1.25  97/06/04  11:58:14  esb
fixed a parsing bug, added timeouts for connections and a GC on cache flush.

Revision 1.24  97/04/22  11:34:06  esb
added a cache.

Revision 1.23  97/04/15  17:58:08  esb
added support for HTTP/1.1.

Revision 1.22  97/03/10  18:46:24  esb
removed use of Chris Okasaki's parser.

Revision 1.21  1996/10/31  15:20:30  cstone
Allows leading blank lines in requests

Revision 1.20  96/10/30  15:13:36  cstone
*** empty log message ***

Revision 1.19  1996/10/30  09:53:19  cstone
better logging

Revision 1.18  96/09/19  17:48:36  esb
fixed syntax error.

Revision 1.17  1996/09/17  16:13:23  cstone
Fixed HEAD requests which require default_document
to be appended to the given URL.

Revision 1.16  1996/09/17  15:56:43  cline
fixed stats in inetd_handler

Revision 1.15  1996/06/07  20:24:00  cline
added inetd service interface

Revision 1.14  1996/05/16  20:01:07  cstone
Sends 32K chunks

Revision 1.13  1996/05/13  16:14:25  cstone
Added server statistics

Revision 1.12  1996/03/02  18:17:10  cstone
Updated to new foxnet

Revision 1.11  1995/11/12  16:51:26  esb
changed an error message.

Revision 1.10  1995/10/17  00:01:00  cstone
Enabled logging and checkpointing

Revision 1.9  1995/10/05  18:53:56  cstone
Added comments

Revision 1.8  1995/10/04  18:57:24  cstone
Changed lstat to stat

Revision 1.7  1995/10/04  13:50:04  cstone
Fixed cgi-bin prefix.

Revision 1.6  1995/10/03  22:20:27  cstone
*** empty log message ***

Revision 1.5  1995/10/03  21:40:38  cstone
Fixed redirections

Revision 1.4  1995/09/29  15:27:20  cstone
New-protocol-signature httpd

Revision 1.3  1995/03/08  20:22:45  cstone
version for public access.

Revision 1.1  1995/01/25  22:43:25  cstone
Initial revision

 
		1.	functor Httpd
*)

functor Httpd (structure B: FOX_BASIS
               structure Lower: PROTOCOL
               val hostname: string
	       val server_pattern: Lower.Pattern.T
	       val setup: Lower.Setup.T
               val default_document: string
               val log_file: string
               val checkpoint_file: string
	       val file_cache_size: int	(* 0 for no caching, size in bytes *)
	       val external_fun : (string * string) list -> string): HTTPD =

 struct
 
  structure Trace = Trace (structure V = B.V
			   val debug_level = NONE
			   val module_name = "httpd.fun"
			   val makestring = fn _ => NONE)

  val server_starttime     = ref (B.V.Time.zeroTime)
  val server_bytesout      = ref (Word64.fromInt 0)
  val server_long_version  = "FoxNet Httpd, Version 2.01, October 30, 1996"
  val server_short_version = "FoxNet-Httpd/2.01"
  val server_accesses      = ref 0


(*
		2.	translation routines
*)

  val explode = B.V.String.explode
  val implode = B.V.String.implode
  val ord     = B.V.Char.ord
  val chr     = B.V.Char.chr
  val ordof   = B.V.String.ordof
  val size    = B.V.String.length
  val tokenize = B.V.String.tokens B.V.Char.isSpace

  val ord_zero   = ord #"0"
  val ord_lowera = ord #"a"
  val ord_uppera = ord #"A"

  (* Convert a character representing a hexidecimal digit to its
     integer equivalent *)
  fun hexchar_to_dec c =
     if (c >= #"0" andalso c <= #"9") then
 	ord c - ord_zero
     else if (c >= #"A" andalso c <= #"F") then
        10 + (ord c) - ord_uppera
     else if (c >= #"a" andalso c <= #"f") then
        10 + (ord c) - ord_lowera
     else
        0

  (* Translate escaped character sequences (% followed by
     a two-hex-digit character code) to a single character. 
     
     Reference: RFC 1738  Uniform Resource Locators, Section 2.2
   *)
  fun unescape' [] = []
    | unescape' (#"%" :: d1 :: d2 :: xs) =
        chr (16 * (hexchar_to_dec d1) + (hexchar_to_dec d2)) :: unescape' xs
    | unescape' (x::xs) = x :: (unescape' xs)

  val unescape = implode o unescape' o explode

  (* Though lines in an HTTP request are defined as being terminated
     by a carriage-return,line-feed sequence, a tolerant server (such
     as this one) should accept a broader range of endline sequences.
     Here we turn all recognized sequences into a simple LF ("\n", chr 10).
     We recognize :  CRLF, LF, or CR.

     Reference:  HTTP 1.0 Internet-Draft 
                 http://www.ics.uci.edu/pub/ietf/http/ *)
  fun fix_endlines (#"\^M" :: #"\n" :: xs) = #"\n" :: (fix_endlines xs)
    | fix_endlines (#"\^M" :: xs)          = #"\n" :: (fix_endlines xs)
    | fix_endlines (x :: xs)               = x :: (fix_endlines xs)
    | fix_endlines []                      = []

  (* Form data has an additional translation: spaces are replaced by
     +'s and true +'s are escaped via the %dd mechanism. The following
     function undoes the space-to-plus translation, and must be done
     before the escaped pluses are unescaped.

     Reference: http://www.w3.org/hypertext/WWW/MarkUp/html-spec/
                       html-spec_8.html#SEC88

                or the full HTML 2.0 document:

                http://www.w3.org/hypertext/WWW/MarkUp/html-spec/
                       html-spec_toc.html

   *)
  fun plus_to_space []           = []
    | plus_to_space (#"+" :: xs) = #" " :: (plus_to_space xs)
    | plus_to_space (x :: xs)    = x  :: (plus_to_space xs)

(*
		3.	Act upon requests
*)

  datatype action = 
    SERVE | RETRY | REDIRECT | EXEC_CGI 

  (* for now, 
        $ = name without the prefix, 
       $$ = full name
      $$$ = full name before unencoding
   *)
  val default_prefix_actions =
    [("/~",       (SERVE,"/afs/cs.cmu.edu/project/fox/mosaic/people/$")),
     ("/afs/cs/", (RETRY,"/afs/cs.cmu.edu/$")),
     ("/afs/cs.cmu.edu/project/fox/mosaic/cgi-bin/", (EXEC_CGI, "$")),
     ("/afs/cs.cmu.edu/project/fox/mosaic/", (SERVE, "$$")),
     ("/afs/cs.cmu.edu/", (REDIRECT, "http://www.cs.cmu.edu$$$")),
     ("/",        (RETRY, "/afs/cs.cmu.edu/project/fox/mosaic/$"))]

  val prefix_actions = ref default_prefix_actions

  fun insert_file (long_suffix, full_name, encoded_name) =
    let
      fun loop [] = []
        | loop (#"$" :: #"$" :: #"$" :: cs) = (explode encoded_name) @ cs
        | loop (#"$" :: #"$" :: cs) = (explode full_name) @ cs
        | loop (#"$" :: cs) = (explode long_suffix) @ cs
        | loop (c :: cs) = c :: (loop cs)
    in
      implode o loop o explode
    end

  fun prefix_of s t =
    let fun loop [] ys      = SOME (implode ys)
	  | loop _  []      = NONE
	  | loop (x::xs) (y::ys) = 
	    if (x <> y) then NONE else loop xs ys
    in
       loop (explode s) (explode t)
    end

  (* forbid * and /../ from appearing in filenames *)
  val insecure =
      let fun insecure' (#"*" :: xs) = true
            | insecure' (#"/" :: #"." :: #"." :: #"/" :: xs) = true
            | insecure' (x :: xs) = insecure' xs
            | insecure' [] = false
      in
         insecure' o explode
      end

  exception File_Not_Found of string

  (* We assume that we've already split out the _filename_ from the
     params and searchpath parts of the URL.  Or else we assume there
     are none. *)
  fun respond {serve, redirect, exec} encoded_name = 
    let
      val full_name = unescape encoded_name

      fun loop full_name [] =
           (Trace.local_print ("unable to find an action for " ^ full_name);
	    raise (File_Not_Found full_name))
	| loop full_name ((prefix, (action, str)) :: xs) = 
  	     (case (prefix_of prefix full_name) of
		NONE => loop full_name xs
	      | SOME long_suffix => 
		  let
		    val action_string = 
		      insert_file (long_suffix, full_name, encoded_name) str
		  in
		    case action of
		      RETRY => 
			loop action_string (! prefix_actions)
		    | SERVE =>
			serve action_string
		    | REDIRECT =>
			redirect action_string
		    | EXEC_CGI =>
			exec action_string
		  end)
    in
      loop full_name (! prefix_actions)
    end

(*
		4.	"server-side includes" (sort of)
*)


  type imported_fun = unit -> string
  type 'a env       = (string * 'a) list ref
  val  exp_funs     = ref [] : imported_fun env

  exception HTTPD_Lookup
  fun empty_env () = ref [] : imported_fun env
  fun lookup (env: 'a env) n = 
    let fun lookup' [] n = raise HTTPD_Lookup
          | lookup' ((x,f)::xs) n = if (x = n) then f else lookup' xs n
    in
      lookup' (!env) n
    end
  fun extend env n f = env := (n,f)::(!env)

  fun exec (name:string) : string = (lookup exp_funs name) ()

  val install_fn = extend exp_funs

  local
   val command_match = "<!--#exec cmd=\""
   val match_length = B.V.String.length command_match
   val command_end = "\"-->"
   val end_skip = B.V.String.length command_end

   fun match_start ([], _) = true
     | match_start (_, []) = false
     | match_start (f1 :: r1, f2 :: r2) =
        f1 = f2 andalso match_start (r1, r2)

   fun match_lists ([], [], count) = SOME count
     | match_lists (sub, [], _) = NONE
     | match_lists (sub, super as (first :: rest), count) =
        if length sub > length super then NONE
	else if match_start (sub, super) then SOME count
	else match_lists (sub, rest, count + 1)

   fun strings_index (substring, superstring, start) =
        match_lists (B.V.String.explode substring,
		     ((List.drop (B.V.String.explode superstring, start))
		      handle Subscript => []),
		     start)

  in
   fun replace_execs s =
        case strings_index (command_match, s, 0) of
           NONE => s
	 | SOME start => 
	    case strings_index (command_end, s, start + match_length) of
	       NONE => s
	     | SOME end_index => 
		let fun sub (start, finish) =
		         B.V.String.substring (s, start, finish - start)
		    val front = sub (0, start)
		    val command = sub (start + match_length, end_index)
		    val back = sub (end_index + end_skip, B.V.String.length s)
		in Trace.debug_print (fn _ =>
				      "string " ^ s ^ ", front " ^ front ^
				      ", command " ^ command ^
				      ", back " ^ back);
		   front ^
		   ((exec command) handle HTTPD_Lookup => "???") ^
		   (replace_execs back)
		end
   end

(*
		5.	parsing functions
*)

  infix X
  fun f X g = fn (x,y) => (f x, g y)

  (* extract the extension from a filename *)
  fun split_extension str =
      let val dot = #"."
	  fun ext_list [] ext = ([], ext)
	    | ext_list (x::xs) ext = 
	         if (x = dot) then (xs, ext) 
                 else ext_list xs (x::ext)

	  val base_and_ext = ext_list (rev (explode str)) []
      in 
	  ((implode o rev) X implode) base_and_ext
      end

  fun extensions str =
    (case split_extension str of
       ("", ext) => [ext]
     | (base, ext) => ext :: (extensions base))

  fun should_parse filename =
     (case (split_extension filename) of
         (_,"shtml") => true
       | _           => false)

  (* Reference:  HTTP/1.0 Internet-Draft v3.0, September 4, 1995 *)
  val default_compress_extensions =
    [("Z",  "x-compress"),
     ("gz", "x-gzip")]

  val compress_extensions = ref default_compress_extensions

  (* Reference: RFC 1700 Assigned Numbers (or its successor RFC).
     also ftp://ftp.isi.edu/in-notes/iana/assignments/media-types
     for an up-to-date listing *)
  val default_file_extensions =
    [("ps",    "application/postscript"),
     ("dvi",   "application/x-dvi"),
     ("pdf",   "application/pdf"),
     ("txt",   "text/plain"),
     ("html",  "text/html"),
     ("htm",   "text/html"),
     ("shtml", "text/html"),
     ("gif",   "image/gif"),
     ("xbm",   "image/x-xbm"),
     ("jpe",   "image/jpeg"),
     ("jpg",   "image/jpeg"),
     ("jpeg",  "image/jpeg"),
     ("tif",   "image/tiff"),
     ("tiff",  "image/tiff"),
     ("mpeg",  "video/mpeg")]

  val file_extensions = ref default_file_extensions

  fun filename_to_filetype filename = 
    let
      fun loop [] ext = "text/plain"
	| loop ((e,m)::es) ext = if (e = ext) then m else loop es ext

      fun loop' [] ext = NONE
	| loop' ((e,c)::es) ext = if (e = ext) then SOME c else loop' es ext

      val extensions = extensions filename
    in
      case loop' (! compress_extensions) (hd extensions) of
        SOME compress =>
	  (SOME compress, loop (! file_extensions) (hd (tl extensions)))
      | NONE =>
	  (NONE, loop (! file_extensions) (hd extensions))
    end
	  
  val html_type = (NONE, "text/html")

   (* Split input from a form into key/value string pairs *)
   val split_input =
      let fun splitoff_pair l []           = [rev l]
            | splitoff_pair l (#"&" :: xs) = (rev l) :: (splitoff_pair [] xs)
            | splitoff_pair l (x :: xs)    = splitoff_pair (x::l) xs

          fun splitoff_key l []           = (rev l, [])
            | splitoff_key l (#"=" :: xs) = (rev l, xs)
            | splitoff_key l (x :: xs)    = splitoff_key (x::l) xs

          val decode_list = 
	    implode o fix_endlines o unescape' o plus_to_space
          val decode_pair = decode_list X decode_list
      in
          ((map decode_pair) o 
           (map (splitoff_key [])) o 
           (splitoff_pair []) o 
           explode)
      end

  (* Returns filename & searchpath (neither decoded) *)
  fun parse_filename filename =
    let 
      val implode_pair = implode X implode
      fun loop [] prefix = implode_pair (rev prefix, [])
        | loop (#"?"::spath) prefix = implode_pair (rev prefix, spath)
        | loop (c :: cs) prefix = loop cs (c::prefix)
    in
      loop (explode filename) []
    end

  exception Content_Length_Missing
  fun get_content_length get_line =
    let
      fun skip_headers () =
	  (case (tokenize (get_line ())) of
	       [] => ()
	     | _ => skip_headers ())
	       
      fun loop () = 
	  (case (tokenize (B.V.String.to_lower (get_line ()))) of
	       [] => raise Content_Length_Missing
	     | ["content-length:", len_string] => 
		   (case (explode len_string) of
			(#":" :: xs) => B.V.String.string_to_int (implode xs)
		      | _ => B.V.String.string_to_int len_string)
	     | ["content-length", ":", len_string] =>
			B.V.String.string_to_int len_string
	     | _ => loop ())
    in (loop ()) before (skip_headers ())
    end

(*
		6.	packet-parsing
*)

  structure ESL = Extern_String_Length (structure In = Lower.Incoming
                                        structure Out = Lower.Incoming
                                        structure V = B.V)
  val w0 = Word.fromInt 0

  fun incoming_to_charlist pkt = 
       explode (#1 ((ESL.unmarshal ((pkt, Lower.Incoming.size pkt), w0))))
  
  fun string_to_outgoing s =
       let fun create [] = NONE
	     | create (first :: rest) =
	        SOME (Word8.fromInt (B.V.Char.ord first), rest)
	   val w8 = Word_Array.W8.U_Big.F.new create (String.explode s)
	   val string_size = Word64.fromInt (size s)
       in server_bytesout := Word64.+ (!server_bytesout, string_size);
          Lower.Outgoing.new (Word_Array.from8 w8)
       end

(*
		7.	HTTP/HTML output
*)

  val CRLF = "\n"

  fun http_begin_header (number, msg) = 
       let val code = (Int.toString number) ^ " " ^ msg
       in "HTTP/1.0 " ^ code ^ CRLF ^
          "Server: " ^ server_short_version ^ CRLF ^
	  "Date: " ^ B.V.Time.toDate (B.V.Time.now ()) ^ CRLF
       end

  fun http_content (NONE, mimetype) = "Content-type: " ^ mimetype ^ CRLF
    | http_content (SOME encoding, mimetype) =
       "Content-type: " ^ mimetype ^ CRLF ^
       "Content-encoding: " ^ encoding ^ CRLF

  fun http_uri uri = "URI: <" ^ uri ^ ">" ^ CRLF ^
                     "Location: " ^ uri ^ CRLF

  fun http_end_header () = CRLF

  fun http_ok_header filename_opt =
       string_to_outgoing ((http_begin_header (200, "OK")) ^
			   (case filename_opt of
			       SOME filename =>
				http_content (filename_to_filetype filename)
			     | NONE =>
				http_content html_type) ^ (http_end_header ()))

  fun html_message title bodytext =
       string_to_outgoing ("<HTML><HEAD><TITLE>" ^ title ^
			   "</TITLE><P></HEAD>\n" ^
			   "<BODY><H1>" ^ title ^ "</H1>\n" ^
			   bodytext ^ "<P></BODY></HTML>\n")

  fun http_redirect_header new_url =
       string_to_outgoing ((http_begin_header(301, "Moved Permanently")) ^ 
			   (http_content html_type) ^
			   (http_uri new_url) ^
			   (http_end_header ()))

  fun html_redirect_body new_url =
       (html_message "301 Moved Permanently" 
        ("Moved Permanently.  The file can be reached via <P><A href=\"" ^
	 new_url ^ "\">" ^ new_url ^ "</A>."))

  fun http_badrequest_header () = 
       string_to_outgoing ((http_begin_header(400, "Bad Request")) ^
			   (http_content html_type) ^
			   (http_end_header ()))

  fun html_badrequest_body url_name =
       (html_message "400 Bad Request"
        ("There is a problem with your request, so that the server " ^
         "cannot determine the proper response."))

  fun http_insecure_header () =
       string_to_outgoing ((http_begin_header (403, "Forbidden")) ^
			   (http_content html_type) ^
			   (http_end_header ()))
     
  fun html_insecure_body url_name =
       (html_message "403 Forbidden"
        ("The URL " ^ url_name ^ "is not allowed by this server.")) 
     
  fun http_missing_header () =
       string_to_outgoing ((http_begin_header(404, "Not Found")) ^
			   (http_content html_type) ^
			   (http_end_header ()))

  fun html_missing_body url_name =
       (html_message "404 Not Found" 
        ("The server could not find this URL:<P>" ^ url_name))

  fun http_unimplemented_header () =
       string_to_outgoing ((http_begin_header (501, "Not Implemented")) ^
			   (http_content html_type) ^
			   (http_end_header ()))

  fun html_unimplemented_body () = 
       (html_message "501 Not Implemented"
	("Please mail cstone@cs.cmu.edu with details of what you just did," ^
	 " including the type and version of the client and hardware " ^
	 " you are using, your e-mail address, and any special circumstances" ^
	 " such as whether you are behind a firewall or using a proxy."))

  fun http_version_header () =
       string_to_outgoing ((http_begin_header (505,
					       "HTTP Version not supported")) ^
			   (http_content html_type) ^
			   (http_end_header ()))

  fun html_version_body () = 
       (html_message "505 HTTP Version not supported"
	("Your client uses a version of HTTP that " ^
	 "is not supported by this server."))

(*
		8.	request logging
*)

  (* Function to save the number of bytes served *)
  fun checkpoint s =
       let val dest = TextIO.openOut checkpoint_file
       in TextIO.output (dest, s^"\n");
	  TextIO.closeOut dest
       end
      handle _ => ()

  (* Function to restore the number of bytes served *)
  fun read_checkpoint () =
       let val src = TextIO.openIn checkpoint_file
	   val data = TextIO.inputLine src
	   val _ = TextIO.closeIn src
           val ord0 = ord #"0"
           val ord9 = ord #"9"

	   fun toWord (nil, a) = a
	     | toWord (d::ds, a) =
	        let val ascii = ord d
	         in if ascii >= ord0 andalso ascii <= ord9 then
		     toWord
		     (ds,
		      Word64.+ (Word64.* (Word64.fromInt 10, a),
				Word64.fromInt (ascii - ord0)))
		    else
		     a
		end
       in toWord (explode data, Word64.fromInt 0)
       end
      handle _ => Word64.fromInt 0

  fun write_to_log s =
       let val dest = TextIO.openAppend log_file
       in TextIO.output (dest, s);
          TextIO.closeOut dest;
	  B.V.Print.print s
       end
      handle _ => ()

  fun log (log_entry, checkpoint_data) =
       (write_to_log (B.V.Time.toDate (B.V.Time.now ()) ^ ": " ^ 
		      log_entry);
	checkpoint checkpoint_data)

  fun log_header () = write_to_log (server_long_version ^ "\n")

(*
		9.	internal function process_request
*)


  fun file_exists filename =
       (TextIO.closeIn (TextIO.openIn filename); true)
       handle _ => false

  fun is_file_readable s = 
       Posix.FileSys.ST.isReg (Posix.FileSys.stat s)
       handle _ => false

  val error = Trace.local_print

  local
   fun str_eq (a: string, b) = a = b
   fun str_hash a =
        B.V.List.fold Word.+
	              (B.V.List.map (Word.fromInt o B.V.Char.ord)
	                            (B.V.String.explode a))
		      0w0

   fun data_space data =
        Word.toInt (B.V.List.fold Word.+
		                  (B.V.List.map Lower.Outgoing.size data) 0w0)

   structure Cache = Single_Cache (structure V = B.V
				   type key = string
				   type value = Lower.Outgoing.T
				   val max_elements = 2039 (* prime number *)
				   val max_space = file_cache_size
				   val eq = str_eq
				   val hash = str_hash
				   val debug_level = NONE)

   val file_size_limit = file_cache_size div 10

   fun send_file send filename =
        let val src = TextIO.openIn filename
	    val send_str = send o string_to_outgoing
	    fun loop acc = 
	         let val next = TextIO.inputN (src, 128 * 1024)
	             val out = string_to_outgoing next
		 in if next <> "" then
		     (send out;
		      loop (out :: acc))
		    else		(* done sending, put into cache *)
		     let val space = data_space acc
		         val list = B.V.List.fold Lower.Outgoing.join
				                  (rev acc) out
		     in if space <= 10000 orelse space <= file_size_limit then
			 ((Cache.add (filename, list, space))
			  handle _ => ())
			else ()		(* too large to put into cache *)
		     end
		 end
	    fun parse_loop n str =
	         if TextIO.endOfStream src then
		  if n <= 0 then () else send_str str
		 else if n >= 1024 then (send_str str; parse_loop 0 "")
		 else
		  let val line = TextIO.inputLine src
		  in parse_loop (n + B.V.String.length line)
		                (str ^ (replace_execs line))
		  end
	in ((if (should_parse filename) then (parse_loop 0 "")
	     else
	      case Cache.look filename of
	         SOME data =>
		  (Trace.local_print "sending cached data";
		   send data)
	       | NONE => loop [])
	    handle x =>
	            (Trace.print_handled (x, SOME ("Send of file " ^
						   filename));
		     error ("Send of file " ^ filename ^ " aborted.\n")));
	   ((TextIO.closeIn src) handle _ => ())
	end

  in
   fun serve_file9 send redirect filename = 
        if (insecure filename) then
	 send (html_insecure_body filename)
	else if (is_file_readable filename) then
	 send_file send filename
	else if (is_file_readable (filename ^ default_document)) then
	 send_file send (filename ^ default_document)
	else if (is_file_readable (filename ^ "/" ^ default_document)) then
	 redirect ("http://" ^ hostname ^ filename ^ "/")
        else
	 send (html_missing_body filename)
    
   fun serve_file send redirect filename =
        if (insecure filename) then
	  (send (http_insecure_header ());
	   send (html_insecure_body filename))
        else if (is_file_readable filename) then
	  (send (http_ok_header (SOME filename));
	   send_file send filename)
        else if (is_file_readable (filename ^ default_document)) then
	  serve_file send redirect (filename ^ default_document)
        else if (is_file_readable (filename ^ "/" ^ default_document)) then
          redirect ("http://" ^ hostname ^ filename ^ "/")
        else
	  (send (http_missing_header ());
	   send (html_missing_body filename))

   fun redirect9 send new_url =
        send (html_redirect_body new_url)

   fun redirect send new_url =
        (send (http_redirect_header new_url);
         send (html_redirect_body new_url))

   fun head_redirect send new_url =
        send (http_redirect_header new_url)

   fun head_serve send filename =
        if (insecure filename) then
	  send (http_insecure_header ())
        else if (is_file_readable filename) then
	  send (http_ok_header (SOME filename))
        else if (is_file_readable (filename ^ default_document)) then
	  head_serve send (filename ^ default_document)
        else if (is_file_readable (filename ^ "/" ^ default_document)) then
          head_redirect send ("http://" ^ hostname ^ filename ^ "/")
        else
	  send (http_missing_header ())

   fun bad_request9 send filename = 
        send (html_badrequest_body filename)

   fun bad_request send filename =
        (send (http_badrequest_header ());
         send (html_badrequest_body filename))

   fun exec send args f fname =
        (send (http_ok_header NONE);
         send (f args))

   val flush_cache = Cache.flush

   fun cache_makestring () = Cache.makestring (fn (k, _) => k, "\n")
  end
	
  val unescape_pair = unescape X unescape

  val count_for_gc = ref 0
  fun process_request (args as (connection_name, Lower.C {send, abort, ...}, 
				quit_pipe, get_line, get_chars)) =
       let val this_request = get_line ()
	   fun valid_http "HTTP/1.0" = true
	     | valid_http "HTTP/1.1" = true
	     | valid_http http =
	        (log (("UNRECOGNIZED HTTP VERSION:" ^ http ^ 
		       " >>> " ^ connection_name ^ "\n"), 
		      Word64.toString (! server_bytesout));
		 send (http_version_header ());
		 send (html_version_body ());
		 false)
       in (server_accesses := ! server_accesses + 1;
	   case tokenize this_request of
	      [] =>
	       (log (("EMPTY LINE: '" ^ this_request ^ 
		      "' >>> " ^ connection_name ^ "\n"), 
		     Word64.toString (! server_bytesout));
		B.Scheduler.sleep 0;
		server_accesses := ! server_accesses - 1;
		process_request args)
	    | ["GET", filename'] => 
	       let val (filename, searchpath) = 
		             unescape_pair (parse_filename filename')
	       in log (("GET9 " ^ filename ^ " >>> " ^ connection_name ^ "\n"),
		       Word64.toString (!server_bytesout));
		  respond {serve = serve_file9 send (redirect9 send), 
			   redirect = redirect9 send,
			   exec = bad_request9 send} filename'
	       end
	    | ["GET", filename', http] =>
	       if valid_http http then
	        let val (filename, searchpath) = 
	                      unescape_pair (parse_filename filename')
	        in log (("GET  " ^ filename ^ " >>> " ^ connection_name ^
			 "\n"),
		        Word64.toString (!server_bytesout));
		    respond {serve = serve_file send (redirect send),
			     redirect = redirect send,
			     exec = bad_request send} filename'
	        end
	       else ()
	    | ["HEAD", filename', http] =>
	       if valid_http http then
	        let val (filename, searchpath) = 
		              unescape_pair (parse_filename filename')
	        in log (("HEAD " ^ filename ^ " >>> " ^
			 connection_name ^ "\n"),
		        Word64.toString (!server_bytesout));
		   respond {serve = head_serve send, 
			    redirect = head_redirect send,
			    exec = bad_request send} filename
	        end
	       else ()
	    | ["POST", filename', http] =>
	       if valid_http http then
	        let val (filename, searchpath) = 
	                      unescape_pair (parse_filename filename')
		    val len = get_content_length get_line
		    val args = split_input (get_chars len)
	        in log (("POST " ^ filename ^ " >>> " ^
			 connection_name ^ "\n"),
		        Word64.toString (! server_bytesout));
		    respond {serve = bad_request send,
			     redirect = redirect send, 
			     exec = exec send args
			                 (string_to_outgoing o external_fun)}
		            filename
	        end
	       else ()
	    | ["FLUSH"] =>
	       (log (("CACHE FLUSH AND GC\n from " ^ connection_name ^ "\n"), 
		     Word64.toString (! server_bytesout));
	        flush_cache ();
		SMLofNJ.Internals.GC.doGC 0;
		SMLofNJ.Internals.GC.doGC 0)
	    | ["QUIT"] => B.Pipe.enqueue (quit_pipe, ())
	    | _ => 
	       (log (("UNRECOGNIZED REQUEST:\n" ^ this_request ^ 
		      " >>> " ^ connection_name ^ "\n"), 
		     Word64.toString (! server_bytesout));
		send (http_unimplemented_header ());
		send (html_unimplemented_body ()));
	   if ! count_for_gc >= 100 then
	    (SMLofNJ.Internals.GC.doGC 0;
	     SMLofNJ.Internals.GC.doGC 0;
	     Trace.local_print ("GC done, cache contains:\n" ^
				cache_makestring ());
	     count_for_gc := 0)
	   else
	    (if ! count_for_gc mod 10 = 0 then
	      Trace.local_print ("GC count " ^ Int.toString (! count_for_gc))
	     else
	      Trace.trace_print (fn _ =>
				 "GC count " ^ Int.toString (! count_for_gc));
	     count_for_gc := ! count_for_gc + 1))
       end

  fun connection_timeout (done, conn_key, packet_pipe) () =
       (B.Scheduler.sleep 300000;
	if ! done then ()
	else 
	 (Trace.local_print ("timing out connection " ^
			     Lower.Connection_Key.makestring conn_key ^
			     " after 300 seconds");
	  B.Pipe.enqueue (packet_pipe, NONE);
	  (* check again (forever) to make sure it is really done *)
	  connection_timeout (done, conn_key, packet_pipe) ()))

  exception No_Output_Available
  
  fun httpd_listen_handler quit_pipe conn_key = 
       let val _ = Trace.trace_print (fn _ => "httpd_listen_handler entered")
	   val packet_pipe = B.Pipe.new () : char list option B.Pipe.T
	   val char_stream = ref [] : char list ref
	   fun get_line () =
	        let fun loop (#"\^M" :: #"\n" :: cs) accum =
		         (char_stream := cs;
			  implode (rev accum))
		      | loop (#"\n" :: cs) accum =
			 (char_stream := cs;
			  implode (rev accum))
		      | loop (c :: cs) accum = loop cs (c :: accum)
		      | loop nil       nil = 
			 (case B.Pipe.dequeue packet_pipe of
			     NONE => raise No_Output_Available
			   | SOME p => loop p nil)
		      | loop nil       accum =
			 (case B.Pipe.dequeue packet_pipe of
			     NONE =>
			      (char_stream := nil;
			       B.Pipe.enqueue (packet_pipe, NONE);
			       implode (rev accum))
			   | SOME p => loop p accum)
		    val line = loop (! char_stream) []
		in print (line ^ "\n");
		   line
		end
	   fun get_chars (n : int) =
	        let fun loop _ 0 accum = implode (rev accum)
		      | loop (c :: cs) n accum = loop cs (n - 1) (c :: accum)
		      | loop nil n accum = 
		         (case B.Pipe.dequeue packet_pipe of
			     NONE => raise No_Output_Available
			   | SOME p => loop p n accum)
		in loop (! char_stream) n []
		end
	   fun data_handler (_, packet) = 
	        (Trace.local_print ("received data: " ^
				    B.V.String.implode (incoming_to_charlist
							packet) ^
				    " for " ^
				    Lower.Connection_Key.makestring conn_key);
		 B.Pipe.enqueue (packet_pipe,
				 SOME (incoming_to_charlist packet)))
	   fun status_handler (conn, status) =
	        (Trace.local_print ("received status " ^
				    Lower.Status.makestring status ^
				    " for " ^
				    Lower.Connection_Key.makestring conn_key);
		 B.Pipe.enqueue (packet_pipe, NONE))
	   val done = ref false
	   fun connection_handler connection =
	        ((Trace.trace_print (fn _ =>
				     "httpd connection_handler entered");
		  process_request (Lower.Connection_Key.makestring conn_key, 
				   connection, quit_pipe, get_line, get_chars);
		  done := true;
	          Trace.trace_print (fn _ =>
				     "httpd connection_handler completed"))
		 handle No_Output_Available =>
		         (Trace.local_print ("unfinished command from " ^
					     Lower.Connection_Key.makestring
					     conn_key);
			  done := true))
       in (* B.Scheduler.fork (connection_timeout (done, conn_key, packet_pipe)); *)
	  {connection_handler = connection_handler,
	   data_handler = data_handler,
	   status_handler = status_handler}
       end

  fun httpd_session_fun (Lower.S{connect, listen, extension}) =
       let val quit_pipe = B.Pipe.new () : unit B.Pipe.T
           val _ = print "httpd_session_fun entered\n"
       in listen (server_pattern, Lower.H (httpd_listen_handler quit_pipe),
		  Lower.Count.Unlimited);
	  B.Pipe.dequeue quit_pipe
       end
      handle _ => (error "exception caught in httpd_session_fun"; ())

(*
		10.	function install
*)

  fun run_httpd () = 
       (server_starttime := B.V.Time.now ();
        log_header ();
	server_bytesout := read_checkpoint ();
	Lower.session (setup, httpd_session_fun));

(*
		11.	functions for reporting statistics
*)

  fun uptime () = 
       B.V.Time.toSeconds (B.V.Time.- (B.V.Time.now (), ! server_starttime))

  fun bytesout () = ! server_bytesout

  fun accesses () = ! server_accesses

(*
		12.	inetd_service definition
*)

  structure Transport = Lower

  local
   val first_time = ref true

   fun handler arguments =
    (if ! first_time then
      (first_time := false;
       server_starttime := B.V.Time.now ();
       log_header ();
       server_bytesout := read_checkpoint ())
     else ();
      httpd_listen_handler (B.Pipe.new ()) arguments)
  in
   val inetd_handler = Lower.H handler
  end
 end
