(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Nick Haines  (Nick.Haines@cs.cmu.edu)
	Edoardo Biagioni <Edoardo.Biagioni@cs.cmu.edu>
        Ken Cline    (Ken.Cline@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	This module contains functions for converting data
        between strings, Bytearrays, and ubyte1 flat_vectors.

	Additionally, the function "flatten" views any fully
        flat (i.e. containing no boxed items) locative as a
        ubyte1 flat_vector locative.

	No attempt has been made to make these function efficient.


		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Convert
	2.	fun string2flat_vector
	3.	fun flat_vector2string
	4.	fun bytearray2flat_vector
	5.	fun flat_vector2bytearray
	6.	fun flatten


		iii.	RCS Log
	
$Log: convert.fun,v $
Revision 1.3  1993/12/21  15:26:27  cline
Cleanup following code review.

Revision 1.2  1993/12/15  15:32:35  cline
Cleaned up table of contents..

Revision 1.1  1993/12/14  21:44:34  cline
Initial revision


		1.	functor Convert

*)

local
  structure L  = Locative
  structure FV = Flat_Vector
in
  functor Convert() : CONVERT =
  struct

    exception flatten_box

(*
	2.	fun string2flat_vector
 *)
    fun string2flat_vector s =
      let
        fun loop k =
          if k = (size s)
            then []
            else (Byte1.from_int (ordof (s, k))) :: (loop (k+1))
      in
        FV.flat_vector (loop 0)
    end

(*
	3.	fun flat_vector2string
 *)
  fun flat_vector2string v =
    let
      fun loop k =
        if k = (FV.flat_length v)
          then ""
          else (chr (Byte1.to_int (FV.flat_sub (v,k)))) ^ (loop (k+1))
    in
      loop 0
    end

(*
	4.	fun bytearray2flat_vector
 *)
  fun bytearray2flat_vector b =
    let
      val length = ByteArray.length b
      val v = locative (FV.flat_vector_const (1u0, length))
      fun loop k =
        if k >= length
         then v
         else (L.locset
                (L.locsub (v, k),
                 Byte1.from_int (ByteArray.sub (b, k))); loop (k+1))
      val result = loop 0;
    in
      L.locget (loop 0)
    end

(*
	5.	fun flat_vector2bytearray
 *)
  fun flat_vector2bytearray (v:ubyte1 FV.flat_vector) =
    let
      val length = FV.flat_length v
      val b = ByteArray.array (length, 0)
      fun loop k =
        if k >= length
         then b
         else (ByteArray.update
                (b, k, Byte1.to_int (FV.flat_sub (v, k))); loop (k+1))
    in
      loop 0
    end

(*
	6.	fun flatten
 *)
  fun flatten flat_loc =
    let
      fun vector_fold (f, v, x) =
        let
          fun fold(i,x) = if i < 0 then x else fold (i-1, f(Vector.sub(v,i),x))
        in
          fold(Vector.length v - 1, x)
        end

      fun descriptor_size (desc_vector(n,d)) = n * descriptor_size d
        | descriptor_size (desc_tuple dvector) =
            vector_fold ((fn (d, i) => i + (descriptor_size d)), dvector, 0)
        | descriptor_size desc_byte4 = 4
        | descriptor_size desc_byte2 = 2
        | descriptor_size desc_byte1 = 1
        | descriptor_size desc_box = raise flatten_box
      val n = descriptor_size (L.locdesc flat_loc)
    in
      (L.locview (flat_loc, desc_vector(n, desc_byte1))) :
         ubyte1 FV.flat_vector locative
    end

  end (* functor Convert *)
end (* local *)