(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Nick Haines  (Nick.Haines@cs.cmu.edu)
	Edoardo Biagioni <Edoardo.Biagioni@cs.cmu.edu>
        Ken Cline    (Ken.Cline@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	Support for reading realm information from configuration file.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Krb_Realm
	2.	fun is_whitespace
	3.	fun input_string
	4.	fun skip_whitespace
	5.	fun local_realm
	6.	fun kerberos_servers


		iii.	RCS Log
	
$Log: krbrealm.fun,v $
Revision 1.1  1994/05/20  02:25:06  robby
Initial revision



		1.	functor Krb_Realm

Kerberos realm information is typically kept in the file `/etc/krb.conf'.
 *)


functor Krb_Realm(val config_file        : string) : KRB_REALM =
struct

(*
	2.	fun is_whitespace
 *)
  fun is_whitespace "" = true
    | is_whitespace " " = true
    | is_whitespace "\t" = true
    | is_whitespace "\n" = true
    | is_whitespace "\000" = true
    | is_whitespace _ = false

(*
	3.	fun input_string

	Read a string up to, but not including whitespace, newline, or eof.
 *)
  fun input_string f =
      case lookahead f of
        "" => ""
      | c  => if is_whitespace (input (f, 1))
                then ""
                else c ^ (input_string f)

(*
	4.	fun skip_whitespace
 *)
  fun skip_whitespace f =
      while is_whitespace (lookahead f) do input (f, 1)

(*
	5.	fun local_realm

	Return the local realm name, which is named in the config_file
 *)
  fun local_realm () =
      let
        val line =
            let val config = open_in config_file
            in input_line config before close_in config end
      in
	if line = "" then ""
        else substring (line, 0, size line - 1)
      end

(*
	6.	fun kerberos_servers

	Return the list of hostnames that config_file lists as
        Kerberos servers for the given realm
 *)
  fun kerberos_servers realm =
      let
        val config = open_in config_file

        (* skip the first line, which names the local realm *)
        val _ = input_line config

        fun list_servers () =
            let
              val realm' = input_string    config
              val _      = skip_whitespace config
              val server = input_string    config
              val _      = input_line      config
            in
              if realm = realm' then
                server :: (list_servers ())
              else if not (end_of_stream config) then
                list_servers ()
              else
                []
            end
      in
        (list_servers ())
        before (close_in config)
      end;

end (* functor *)
