(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Robert Findler (Robert.Findler@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	An implementation of the Translate functor.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	datatype Specification
	2.	datatype Exr
	3.	exception Does_Not_Match_Specification
	4.	fun internalize
	5.	fun externalize

		iii.	RCS Log
	
$Log: translate.fun,v $
Revision 1.1  1994/06/27  16:41:25  robby
Initial revision

*)

functor Translate (structure Copy:COPY
		   structure Access:ACCESS
		   structure O:BYTE_ORDERS):TRANSLATE=struct

exception Internal_Error of string
exception Translate_Error
(*
		1.	datatype Specification
*)
datatype Specification = sUbyte4 | sUbyte2 | sUbyte1 | sPString |
  sCString | sString of int | sTuple of Specification list |
  sTagged of (Specification*(Exr->Specification)) |
  sTrailing_Tuple of Specification list | sArray of int*Specification |
  sByteArray of int | sEmpty

(*
		2.	datatype Exr
*)
and Exr = 
  eUbyte4 of Byte4.ubytes | eUbyte2 of Byte2.ubytes | eUbyte1 of Byte1.ubytes |
  eString of string | eTuple of Exr list | eByteArray of ByteArray.bytearray |
  eEmpty

datatype Pos_Exr=Pos_eUbyte4 of ubyte4*int | Pos_eUbyte2 of ubyte2*int |
  Pos_eUbyte1 of ubyte1*int | Pos_eString of string*int*int | 
  Pos_eTuple of Pos_Exr list | Pos_eByteArray of ByteArray.bytearray*int |
  Pos_eEmpty


(*
		3.	exception Does_Not_Match_Specification
*)
exception Does_Not_Match_Specification of Specification*ByteArray.bytearray*int

fun position_exr (eUbyte4 a) p=(Pos_eUbyte4(a,p),p+4)
  | position_exr (eUbyte2 a) p=(Pos_eUbyte2(a,p),p+2)
  | position_exr (eUbyte1 a) p=(Pos_eUbyte1(a,p),p+1)
  | position_exr (eString s) p=
    let val len=Fox_Basis.V.String.length s
    in (Pos_eString(s,p,len),p+len) end
  | position_exr (eTuple r) p=
    let 
      fun helper (r::rs) p'=
	let val (r',p'')=position_exr r p'
	  val (rs',p''')=helper rs p''
	in (r'::rs',p''') end
	| helper [] p=([],p)
      val (list,size)=helper r p
    in
      (Pos_eTuple list,size)
    end
  | position_exr (eByteArray b) p=(Pos_eByteArray (b,p),p+ByteArray.length b)
  | position_exr (eEmpty) p=(Pos_eEmpty,p)

fun externalize_helper (Pos_eUbyte4 (u,p)) a=
  let
    val u=O.B4.to_big u
    val one=Byte4.to_int (Byte4.&&(4ux000000FF,u))
    val two=Byte4.to_int (Byte4.>>((Byte4.&&(4ux0000FF00,u)),8))
    val three=Byte4.to_int (Byte4.>>((Byte4.&&(4ux00FF0000,u)),16))
    val four=Byte4.to_int (Byte4.>>((Byte4.&&(4uxFF000000,u)),24))
  in
    ByteArray.update (a,p,one);
    ByteArray.update (a,(p+1),two);
    ByteArray.update (a,(p+2),three);
    ByteArray.update (a,(p+3),four);
    a
  end
  | externalize_helper (s as Pos_eUbyte2 (u,p)) a=
    let
      val u=O.B2.to_big u
      val one=Byte2.to_int (Byte2.&&(2ux00FF,u))
      val two=Byte2.to_int (Byte2.>>(Byte2.&&(2uxFF00,u),8))
    in
       ByteArray.update (a,p,one);
       ByteArray.update (a,(p+1),two);
       a
    end
  | externalize_helper (Pos_eUbyte1 (u,p)) a=
    (ByteArray.update (a,p,(Byte1.to_int u));
     a)
  | externalize_helper (Pos_eString (s,p,len)) a=
    let fun helper s p_string p_array=
       if (p_string = len) then a else
	 (ByteArray.update (a,p_array,(Fox_Basis.V.String.ordof (s,p_string)));
	  helper s (p_string+1) (p_array+1))
    in
      helper s 0 p
    end
  | externalize_helper (Pos_eTuple []) a=a
  | externalize_helper (Pos_eTuple (r::rs)) a=
    (externalize_helper r a;
     externalize_helper (Pos_eTuple rs) a)
  | externalize_helper (Pos_eByteArray (array,pos)) a=
    let val length=ByteArray.length array
    in
      Copy.copy (array,0,length,array,pos);
      a
    end
  | externalize_helper (Pos_eEmpty) a=a

fun internalize_helper pos sUbyte4 a=
  ((eUbyte4 (O.B4.from_big (Byte4.from_int (ByteArray.sub (a,pos+3))) +
	     (Byte4.<< (Byte4.from_int 
			(ByteArray.sub (a,pos+2)),8)) +
	     (Byte4.<< (Byte4.from_int 
			(ByteArray.sub (a,pos+1)),16)) +
	     (Byte4.<< (Byte4.from_int 
			(ByteArray.sub (a,pos)),24))),
  pos+4) handle ByteArray.Subscript => 
  raise Does_Not_Match_Specification (sUbyte4,a,pos))
  | internalize_helper pos sUbyte2 a=
    ((eUbyte2 (O.B2.from_big(Byte2.from_int (ByteArray.sub (a,(pos+1)))) +
	       (Byte2.<<(Byte2.from_int 
			 (ByteArray.sub (a,pos)),8))),
      pos+2) handle ByteArray.Subscript =>
     raise Does_Not_Match_Specification (sUbyte2,a,pos))
  | internalize_helper pos sUbyte1 a=
    ((eUbyte1 (Byte1.from_int (ByteArray.sub (a,pos))),pos+1)
     handle ByteArray.Subscript => 
       raise Does_Not_Match_Specification (sUbyte1,a,pos))
  | internalize_helper pos sPString a=
    (let val length=ByteArray.sub (a,pos)
    in
       (eString (Access.range_to_string 
		 (a,Access.Length {first=pos+1,length=length})),
	pos+length+1)
    end handle Access.Bad_Range =>
	raise Does_Not_Match_Specification (sPString,a,pos))
  | internalize_helper pos sCString a=
    (let val string=Access.range_to_string
       (a,Access.Test {first=pos,test=(fn x=>x=1u0)})
    in 
       (eString string,pos + 1 + Fox_Basis.V.String.length string)
    end
       handle Access.Bad_Range => 
	 raise Does_Not_Match_Specification (sCString,a,pos))
  | internalize_helper pos (sString length) a=
    (eString (Access.range_to_string 
	      (a,Access.Length {first=pos,length=length})),pos+length)
  | internalize_helper pos (sTagged (spec,find_rest)) a=
    let val (first,pos')=internalize_helper pos spec a
      val spec_rest=find_rest first
      val (rest,pos'')=internalize_helper pos' spec_rest a
    in 
      (eTuple [first,rest],pos'')
    end
  | internalize_helper pos (sTuple t) a=
    let fun helper pos [] a=([],pos)
	  | helper pos (t::ts) a=
	    let val (t',pos')=internalize_helper pos t a
	      val (ts',pos'')=helper pos' ts a
	    in (t'::ts',pos'') end
	val (t',pos')=helper pos t a
    in (eTuple t',pos') end
  | internalize_helper pos (sTrailing_Tuple t) a=
    let val size=ByteArray.length a
      val (exr,pos')=internalize_helper pos (sTuple t) a
    in
      case exr of
	eTuple t'=>
	  (eTuple (t'@[(eByteArray (Copy.copy_create (a,pos',size-pos')))]),
	   size)
      | _ => raise (Internal_Error "internalize_helper sTrailing_Tuple")
    end
  | internalize_helper pos (sArray (size,spec)) a=
    let fun helper pos n a=
             if pos=size then ([],pos) else
	       let val (t,pos')=internalize_helper pos spec a
		 val (ts,pos'')=helper pos' (n+1) a
	       in (t::ts,pos'') end
	val (ts,pos')=helper pos 0 a
    in (eTuple ts,pos') end
  | internalize_helper pos (s as sByteArray size) a=
    ((eByteArray (Copy.copy_create (a,pos,size)),size+pos)
     handle (Copy.Illegal_Copy _) => 
       raise Does_Not_Match_Specification (s,a,pos))
  | internalize_helper pos sEmpty a=(eEmpty,pos)
        

(*
		4.	fun externalize
*)

fun externalize exr=
  let val (positioned_exr,size)=position_exr exr 0
  in externalize_helper positioned_exr (ByteArray.array (size,0)) end

(*
		5.	fun internalize
*)
fun internalize s a=
  let val (ans,size)=internalize_helper 0 s a
  in if size <> (ByteArray.length a) then
       raise Does_Not_Match_Specification (sEmpty,a,size)
     else 
       ans
  end
end
