(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Robert Findler (Robert.Findler@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	The signature for the Translate functor. Translate will take a
	member of the Exr datatype and turn it into a bytearray suitable
	for sending over the wire. Everything will be in big endian format.
	The Exr's will be parsed from left to right.

	Translate will also convert bytearrays into Exr's using the
	Specification datatype. The Does_Not_Match_Specification exception
	will be raised if the bytearray cannot be coerced into the
	Specification specified. The execption carries with it the position
	where the specification failed and the specification that was expected.
	
		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	signature TRANSLATE

		iii.	RCS Log
	
$Log: translate.sig,v $
Revision 1.1  1994/06/27  16:41:27  robby
Initial revision


		1.	signature TRANSLATE
*)
signature TRANSLATE=sig

datatype Specification = 
  sUbyte4 | sUbyte2 | sUbyte1 |                     
  sPString |                                        (*strings with a leading 
						      byte that specifies size*)
  sCString |                                        (*null terminated strings*)
  sString of int |                                  (*string of a given length*)
  sTuple of Specification list |                    (*tuples*)
  sTagged of (Specification*(Exr->Specification)) | (*if part of the
				  		      specification depends on 
						      an earlier part*)
  sTrailing_Tuple of Specification list |           (*if the specification has 
						      some unspecified number
						      of trailing bytes*)
  sArray of int*Specification |                     (*This will return as a 
						      Tuple in the Exr 
                                                      datatype.*)
  sByteArray of int |                               (*makes a copy*)
  sEmpty

  and Exr = 
    eUbyte4 of ubyte4 |
    eUbyte2 of ubyte2 | 
    eUbyte1 of ubyte1 | 
    eString of string |
    eTuple of Exr list |
    eByteArray of ByteArray.bytearray | 
    eEmpty

  (* This exception is provided for use with sTagged and pattern matching.*)
  exception Translate_Error

  (* the part of the specification which fails to match the bytearray,
     the bytearray which doesn't match and the position where the matched
     began to fail. *)
  exception Does_Not_Match_Specification of 
            Specification*ByteArray.bytearray*int

  val externalize : Exr -> ByteArray.bytearray
    
  val internalize : Specification -> ByteArray.bytearray -> Exr

end
