(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

Trivial zwrite application.

In order to send a z'gram to the local host manager, EthDev's
high_priority_filter must be false.  Unfortunately, this setting
interferes with arp, so Udp's use_arp must also be false, which
means the IP resolution table has to contain the ethernet address
for the machine on which zwrite is run.

zwrite can be invoked as follows (using your host's IP number instead of
4ux12345678):

	structure ZW = ZWrite (val zephyr_host = 4ux12345678
			       val do_prints = false);

	ZW.zwrite ("recipient", "sender", "message");


		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor zwrite

		iii.	RCS Log
	
$Log: zwrite.sml,v $
Revision 1.8  1994/07/13  13:56:49  robby
*** empty log message ***

Revision 1.7  94/04/28  20:15:58  cline
minor changes

Revision 1.6  1994/04/28  20:10:31  cline
zwrite now works with low_priority_filter=false and use_arp=false.

Revision 1.5  1994/04/26  20:01:11  esb
minor changes.

Revision 1.4  1994/04/25  17:19:57  cline
update to current foxnet.  Some bug prevents this version from working

Revision 1.3  1993/12/15  15:15:20  cline
removed #s from changelog.

Revision 1.2  1993/10/25  19:39:30  cline
removed .U from Byte[421].U

Revision 1.1  1993/10/20  14:21:58  cline
Initial revision


		1.	functor zwrite
*)

functor ZWrite (structure B:FOX_BASIS
		sharing type Buffer.buf=B.Filter.ethernet_packet_filter
		and type ByteArray.bytearray=B.Format.bytearray
                val do_prints: bool) =
struct

  val ip_over_eth = 2ux800
  val arp_over_eth = 2ux806

  val udp_over_ip = 1ux11

  val zephyr_hm_port = 2u14623 (* debug *)
  val zephyr_hm_port = 2u2104

  structure Scheduler = Coroutine (structure Debug = Fox_Basis.Debug
			           structure V = Fox_Basis.V
				   val do_prints = do_prints)

  structure Dev = BuildEthDev (structure B=B
			       structure Scheduler = Scheduler
                               val protocol_numbers = [ip_over_eth,
                                                       arp_over_eth]
			       val high_priority_filter = true
			       val do_prints = do_prints)


  structure Stack = BuildUdp (structure B=B
			      structure Scheduler = Scheduler
			      structure Device    = Dev.EthDev
			      val ip_over_eth     = ip_over_eth 
			      val udp_over_ip     = udp_over_ip
			      val use_arp         = false
			      val do_prints       = do_prints)

  structure Udp = Stack.Udp
  structure Ip = Stack.Ip

  val (local_ip, local_interface, gateway_ip) = Stack.Resolve.local_info ()

  val header_bytes = 50 (*This is an overestimate*)

  (* recipient is a UDP address *)
  val recipient =
    let val ip_addr = Ip.Address {ip=local_ip, proto=udp_over_ip}
    in
      Udp.Address {remote_peer = ip_addr,
                   local_port  = 2u0,
		   remote_port = zephyr_hm_port}
    end

  fun build_packet (whom, from, text) =
    let

      fun build_unique_packet_id () =
        let
          val System.Timer.TIME {sec, usec} =
                System.Unsafe.CInterface.gettimeofday ();

          fun hex_digit d = chr ((if d>=10 then (ord "A")-10 else (ord "0"))
                                 + d)

          fun do_digits (_, 0) = ""
            | do_digits (i, n) = (do_digits (i div 16, n-1)) ^
                                 (hex_digit (i rem 16))

          fun hexify i = "0x" ^
                         (if (i<0) then (do_digits (~i, 8))
                                   else (do_digits (i, 8)))

        in "0x8002CD22" ^ " " ^ (hexify sec) ^ " " ^ (hexify usec) ^ "\000"
        (* "0x12345678" ^ " " ^ (hexify sec) ^ " " ^ (hexify usec) ^ "\000" *)
        end

      val version_string	  = "ZEPH0.2\000"
      val number_of_header_fields = "0x00000011\000"
      val notice_kind		  = "0x00000001\000"
      val unique_packet_id	  = build_unique_packet_id ()
      val port			  = "0x0000\000"
      val authentication	  = "0x00000000\000"
      val authenticator_length	  = "0x00000000\000"
      val authenticator		  = "\000"
      val class			  = "MESSAGE\000"
      val class_instance	  = "PERSONAL\000"
      val opcode		  = "\000"
      val sender		  = "The FoxNet\000"
      val recipient		  = whom ^ "\000"
      val default_display_format  = "Class $class, Instance $instance:\n" ^
                                    "To: @bold($recipient)\n" ^
                                    "@bold($1) <$sender>\n\n$2\000"
      val checksum		  = "0x00000000\000"
      val fragmentation_count	  = "\000"
      val unique_notice_id	  = unique_packet_id
      val data			  = from ^ "\n\000" ^ text ^ "\n\000"

      fun string_to_bytearray s =
        let
          val a = ByteArray.array (size s, 0)
          fun aux i = if i = size s then ()
		      else (ByteArray.update (a, i, ordof (s, i));
			    aux (i + 1))
        in aux 0;
           a
        end

      val message =
            version_string		^
            number_of_header_fields	^
            notice_kind			^
            unique_packet_id		^
            port			^
            authentication		^
            authenticator_length	^
            authenticator		^
            class			^
            class_instance		^
            opcode			^
            sender			^
            recipient			^
            default_display_format	^
            checksum			^
            fragmentation_count		^
            unique_notice_id		^
            data

    in string_to_bytearray message
    end

  fun zwrite spec = (* spec: (to, from, message): string * string * string *)
    ((Udp.initialize ();
     let val conn = Udp.active_open (recipient, (fn _ => fn _ => ()))
     in Udp.send conn (B.Send_Packet.create ((build_packet spec),
					     header_bytes));
       Udp.close conn
     end;
     Udp.finalize ();
     ()
     handle e => (Udp.finalize(); raise e)))

end
