(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15213-3891

		i.	Abstract

	Access to addressing information for the machine executing the
	program.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	structure Constants
	2.	functor Initialize
	3.	structure Dev
	4.	structure Eth
	5.	structure Ip
	6.	structures Udp and Tcp
	7.	structure Dns

		iii.	RCS Log
	
$Log$

	1.	structure Constants
*)

structure Constants =
  struct

    val ip_protocol = Word16.fromInt 0x800
    val arp_protocol = Word16.fromInt 0x806

    val icmp_protocol = 0wx01:Word8.word
    val tcp_protocol =  0wx06:Word8.word
    val udp_protocol =  0wx11:Word8.word

    val dns_port = Word16.fromInt 53

    structure TCP =
      struct
	val initial_window_size = 4096
	val user_timeout = 10000
      end

    val ip_address = "128.171.10.223"
    val ip_mask = "255.255.255.0"
    val ip_default_router = "128.171.10.1"
    val eth_interface = "eth0"
    val ip_mtu = 1500

    val lo_ip_address = "127.0.0.1"
    val lo_ip_mask = "255.0.0.0"
    val lo_interface = "lo0"
    val lo_ip_mtu = 1500
  end

(*
	2.	functor Initialize
*)

functor Initialize (structure Device: RAW_DEVICE
		    structure V: VENDOR
		    val hostname_to_ip: string -> Ip_Host_Id.T option
		    val interface_number: int) =
  struct

   exception Initialize

   structure Trace = Trace (structure V = V
			    val makestring = fn _ => NONE
			    val module_name = "helloinit.fun"
			    val debug_level = NONE)


   exception Illegal_IP

   fun parse_ip string =
        case hostname_to_ip string of
	   SOME x => x
	 | NONE =>
	    Trace.print_raise (Illegal_IP, SOME ("illegal IP address " ^ string))

(*
	3.	structure Dev
*)
    structure Dev =
      struct
	fun interface () = "eth0"
      end

(*
	4.	structure Eth
*)
    structure Eth =
      struct
	fun local_address () =
	     (Trace.local_print "no local address for ethernet";
	      ())
      end

(*
	5.	structure Ip
*)
    structure Ip =
      struct

	fun setup () =
	  let val local_address = parse_ip Constants.ip_address
	      val default_gateway = parse_ip Constants.ip_default_router
	      val mask = SOME (parse_ip Constants.ip_mask, {serve = false})
	      val mtu = SOME Constants.ip_mtu
	  in [{local_id = local_address,
	       interface = Constants.eth_interface,
	       gateways = [default_gateway],
	       mask = mask,
	       mtu = mtu},
	      {local_id = parse_ip Constants.lo_ip_address,
	       interface = Constants.lo_interface,
	       gateways = [],
	       mask = SOME (parse_ip Constants.lo_ip_mask, {serve = false}),
	       mtu = SOME Constants.lo_ip_mtu}]
	  end
      end

(*
	6.	structures Udp and Tcp
*)
    structure Udp = Ip
    structure Tcp = Ip

(*
	7.	structure Dns
*)
    structure Dns =
      struct
	fun domain_list () =
	     [".", "ics.hawaii.edu"]

	fun servers () = ([]: Word32.word list)
(*
	     [parse_ip "128.171.1.1", parse_ip "128.171.3.13",
	      parse_ip "128.171.10.74"]
*)

	fun setup t = {domain_list = domain_list (),
		       servers = servers (),
		       transport = t}
      end

  end
